//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#ifndef GLBOY_ALU_HH
#define GLBOY_ALU_HH

#include "glBoy.hh"
#include "DWORD.hh"
#include "Registers.hh"

namespace glBoy
{
namespace ALU
{
void load();
extern const uint8_t ParityFlagLookup[256];

inline uint8_t
add(uint8_t a, uint8_t b, Flags& flags, bool carry = false)
{
	uint8_t result(a + b);

	if (carry && flags.C)
	{
		++result;
	}

	flags.byte =
		(result & 0x80) |                       // S  MSB
		((result == 0) ? 0x40 : 0) |            // Z
												// U5  TODO
		((a ^ b ^ result) & 0x10) |             // H
												// U3 TODO

		// If a and b have the same sign, and the result
		// has a different sign, then we overflowed the
		// register.
		// Can never overflow if a and b have different signs. (even with
		// carry, since 127 + -1 + carry == 127).
		// Step 1) Determine whether a and b have the same sign. (a NXOR b)
		// Step 2) Determine whether the result has // a different sign.
		((((~(a ^ b)) & (result ^ a)) & 0x80) >> 5) |// P

												// N
		(((uint16_t(a) + uint16_t(b)) & 0x100) >> 8); // C

	return result;
}

inline uint8_t
adc(uint8_t a, uint8_t b, Flags& flags)
{
	return add(a, b, flags, true);
}

inline DWORD
add(DWORD a, DWORD b, Flags& flags, bool carry = false)
{
	const uint16_t aHost(a.host());
	const uint16_t bHost(b.host());

	uint16_t result(aHost + bHost);

	if (carry && flags.C)
	{
		++result;
	}

	if (carry)
	{
		// Flags only affected for adc.
		flags.S = (result & 0x8000) ? 1 : 0;
		flags.Z = (result == 0);
		flags.P = (((~(aHost ^ bHost)) & (result ^ aHost)) & 0x8000) ? 1 : 0;

	}

	flags.N = 0;
	flags.H = ((aHost ^ bHost ^ result) >> 10) & 0x1;
	flags.C = (((uint32_t(aHost) + uint32_t(bHost)) & 0x10000) >> 16);

	return glBoy::htoz(result);
}

inline DWORD
adc(DWORD a, DWORD b, Flags& flags)
{
	return add(a, b, flags, true);
}

inline uint8_t
sub(uint8_t a, uint8_t b, Flags& flags, bool borrow = false)
{
	uint8_t c = (borrow && flags.C) ? 1 : 0;
	uint8_t result(a - b - c);

	flags.byte =
		(result & 0x80) |                       // S  MSB
		((result == 0) ? 0x40 : 0) |            // Z
												// U5  TODO
		(((a & 0xf) - (b & 0xf)) & 0x10) |      // H
												// U3 TODO

		// Detected signed overflow
		((int16_t(static_cast<int8_t>(a)) -
			int16_t(static_cast<int8_t>(b)) -
			c
			) ==
			static_cast<int8_t>(result) ? 0 : 0x4) | // P

		0x2 |										// N
		(((uint16_t(a) - uint16_t(b)) & 0x100) >> 8); // C

	return result;
}

inline uint8_t
sbc(uint8_t a, uint8_t b, Flags& flags)
{
	return sub(a, b, flags, true);
}

inline DWORD
sub(DWORD a, DWORD b, Flags& flags, bool borrow = false)
{
	const uint16_t aHost(a.host());
	const uint16_t bHost(b.host());

	uint16_t c = (borrow && flags.C) ? 1 : 0;
	uint16_t result(aHost - bHost - c);

	flags.byte =
		((result & 0x8000) >> 8) |        // S  MSB
		((result == 0) ? 0x40 : 0) |            // Z
												// U5  TODO
		((((aHost & 0xf00) - (bHost & 0xf00)) & 0x1000) >> 8) | // H
												// U3 TODO

		// Detected signed overflow
		((int32_t(static_cast<int16_t>(aHost)) -
			int32_t(static_cast<int16_t>(bHost)) -
			c
			) ==
			static_cast<int16_t>(result) ? 0 : 0x4) | // P

		0x2 |										// N
		(((uint32_t(aHost) - uint32_t(bHost)) & 0x10000) >> 16); // C

	return htoz(result);
}

inline DWORD
sbc(DWORD a, DWORD b, Flags& flags)
{
	return sub(a, b, flags, true);
}

inline uint8_t
bitwiseAnd(uint8_t a, uint8_t b, Flags& flags)
{
	uint8_t result(a & b);
	flags.byte =
		(result & 0x80) | // S
		((result == 0) ? 0x40 : 0) |            // Z
				// U5
		0x10 | // H
				// U3
		ParityFlagLookup[result]; // P
			// N
			// C
	return result;
}

inline uint8_t
bitwiseOr(uint8_t a, uint8_t b, Flags& flags)
{
	uint8_t result(a | b);
	flags.byte =
		(result & 0x80) | // S
		((result == 0) ? 0x40 : 0) |            // Z
				// U5
				// H
				// U3
		ParityFlagLookup[result]; // P
			// N
			// C
	return result;
}

inline uint8_t
bitwiseXor(uint8_t a, uint8_t b, Flags& flags)
{
	uint8_t result(a ^ b);
	flags.byte =
		(result & 0x80) | // S
		((result == 0) ? 0x40 : 0) |            // Z
				// U5
				// H
				// U3
		ParityFlagLookup[result]; // P
			// N
			// C
	return result;
}

inline uint8_t
rotateLeft(uint8_t a, Flags& flags, bool throughCarry, bool setAllFlags = false)
{
	uint8_t msb(a >> 7);
	a <<= 1;

	if (throughCarry)
	{
		a |= flags.C;
	}
	else
	{
		a |= msb;
	}
	flags.H = 0;
	flags.N = 0;
	flags.C = msb;

	if (setAllFlags)
	{
		flags.byte |= (a & 0x80);
		flags.Z = (a == 0) ? 1 : 0;
		flags.byte |= ParityFlagLookup[a];
	}
// TODO U3 U5

	return a;
}

inline uint8_t
shiftLeft(uint8_t a, Flags& flags)
{
	uint8_t msb(a >> 7);
	a <<= 1;

	flags.byte =
		(a & 0x80) | // S
		((a == 0) ? 0x40 : 0) | // Z
		// U5
		// H
		// U3
		ParityFlagLookup[a] | // P
		// N
		msb; // C

	return a;
}

inline uint8_t
rotateRight(uint8_t a, Flags& flags, bool throughCarry, bool setAllFlags = false)
{
	uint8_t lsb(a & 0x1);
	a >>= 1;

	if (throughCarry)
	{
		a |= flags.C ? 0x80 : 0;
	}
	else
	{
		a |= lsb ? 0x80 : 0;
	}
	flags.H = 0;
	flags.N = 0;
	flags.C = lsb;

	if (setAllFlags)
	{
		flags.byte |= (a & 0x80);
		flags.Z = (a == 0) ? 1 : 0;
		flags.byte |= ParityFlagLookup[a];
	}
// TODO U3 U5

	return a;
}

inline uint8_t
shiftRight(uint8_t a, Flags& flags, bool arithmatic)
{
	uint8_t lsb(a & 0x1);
	uint8_t msb(a & 0x80);
	a >>= 1;

	// Restore sign bit
	if (arithmatic) a |= msb;

	flags.byte =
		(a & 0x80) | // S
		((a == 0) ? 0x40 : 0) | // Z
		// U5
		// H
		// U3
		ParityFlagLookup[a] | // P
		// N
		lsb; // C

	return a;
}

inline void
RLD(uint8_t& a, uint8_t& m, Flags& f)
{
	// Whoever came up with RLD was on crack.

	uint8_t newM = (a & 0xf) | (m << 4);
	a = (a & 0xF0) | (m >> 4);
	m = newM;

	f.byte =
		(a & 0x80) | // S
		((a == 0) ? 0x40 : 0) | // Z
		// U5
		// H
		// U3
		ParityFlagLookup[a]; // P
		// N
		// C
}

inline void
RRD(uint8_t& a, uint8_t& m, Flags& f)
{
	// Whoever came up with RRD was on crack.

	uint8_t newM = (a << 4) | (m >> 4);
	a = (a & 0xF0) | (m & 0xf);
	m = newM;

	f.byte =
		(a & 0x80) | // S
		((a == 0) ? 0x40 : 0) | // Z
		// U5
		// H
		// U3
		ParityFlagLookup[a]; // P
		// N
		// C
}

} // namespace ALU

} // namespace m80
#endif

