//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#define GL_GLEXT_PROTOTYPES

#include "glBoy.hh"
#include "GameboyGraphics.hh"
#include "Log.hh"
#include "MemoryMap.hh"

#include <GL/gl.h>
#include <GL/glext.h>

#include <SDL/SDL_opengl.h>

#include <cassert>
#include <limits>

#include <sys/time.h>


using namespace glBoy;

namespace
{
	class VRAM : public MemoryMap::Memory
	{
	public:
		VRAM(uint8_t* mem,
			const std::function<void ()>& changed
			) :
			m_mem(mem),
			m_changed(changed)
		{}

		virtual uint8_t read8(uint16_t address) { return m_mem[address]; }
		virtual void write8(uint16_t address, uint8_t value)
		{
			m_mem[address] = value;
			m_changed();
		}

		virtual void copy(uint8_t* dest, uint16_t address, int bytes)
		{
			memcpy(dest, m_mem + address, bytes);
		}

	private:
		uint8_t* m_mem;
		std::function<void ()> m_changed;
	};

}

class GameboyGraphics::GraphicsRegisters : public MemoryMap::Memory
{
public:
	uint8_t LCDC;
	uint8_t STAT;
	uint8_t SCY;
	uint8_t SCX;
	uint8_t LY;
	uint8_t LYC;
	// uint8_t DMA; No need to store this - write-only.
	uint8_t BGP;
	uint8_t OBP0;
	uint8_t OBP1;
	uint8_t WY;
	uint8_t WX;

	GraphicsRegisters(
		Core& core,
		uint8_t* oam,
		const std::function<void ()>& backgroundChanged
		) :
		LCDC(0),
		STAT(0),
		SCY(0),
		SCX(0),
		LY(0),
		LYC(0),
		BGP(0),
		OBP0(0),
		OBP1(0),
		WY(0),
		WX(0),
		m_core(core),
		m_mapping(core.getMemoryMap()),
		m_oam(oam),
		m_backgroundChanged(backgroundChanged)
	{
	}

	virtual uint8_t read8(uint16_t address)
	{
		assert(address < 12);

		switch (address)
		{
			case 0: // LCDC
			{
				return LCDC;
			} break;
			case 1: // STAT
			{
				return STAT;
			} break;
			case 2: // SCY
			{
				return SCY;
			} break;
			case 3: // SCX
			{
				return SCX;
			} break;
			case 4: // LY
			{
				return LY;
			} break;
			case 5: // LYC
			{
				return LYC;
			} break;
			case 6: // DMA
			{
				return 0;
			} break;
			case 7: // BGP
			{
				return BGP;
			} break;
			case 8: // OBP0
			{
				return OBP0;
			} break;
			case 9: // OBP1
			{
				return OBP1;
			} break;
			case 10: // WY
			{
				return WY;
			} break;
			case 11: // WX
			{
				return WX;
			} break;
		}
		return 0; // make g++ happy.
	}

	virtual void write8(uint16_t address, uint8_t value)
	{
		assert(address < 12);

		switch (address)
		{
			case 0: // LCDC
			{
				LCDC = value;
				m_backgroundChanged();
			} break;
			case 1: // STAT
			{
				// bits 0-2 are read only.
				uint8_t tmp(STAT & 0x7);
				STAT = value & 0xF8;
				STAT |= tmp;
			} break;
			case 2: // SCY
			{
				SCY = value;
			} break;
			case 3: // SCX
			{
				SCX = value;
			} break;
			case 4: // LY
			{
				// Readonly.
			} break;
			case 5: // LYC
			{
				LYC = value;
			} break;
			case 6: // DMA
			{
				// Start a DMA transfer to the OAM memory area. This should
				// take 160 microseconds, but there's no real need to simulate
				// this.  CPU may only access zero-page mem during the
				// transfer, and will spin the required number of cycles.
				uint16_t dmaStart(static_cast<uint16_t>(value) << 8);
				m_mapping.copy(m_oam, dmaStart, 0x9F);
			} break;
			case 7: // BGP
			{
				BGP = value;
				m_backgroundChanged();
			} break;
			case 8: // OBP0
			{
				OBP0 = value;
			} break;
			case 9: // OBP1
			{
				OBP1 = value;
			} break;
			case 10: // WY
			{
				WY = value;
			} break;
			case 11: // WX
			{
				WX = value;
			} break;
		}
	}

	private:
		Core& m_core;
		MemoryMap& m_mapping; // For DMA transfers.
		uint8_t* m_oam;
		std::function<void ()> m_backgroundChanged;
};


GameboyGraphics::GameboyGraphics(
	Core& core,
	SDL_Surface* surface,
	PixelScaler scaler,
	ResampleMethod resampler
	) :
	m_core(core),
	m_scaler(scaler),
	m_resampler(resampler),
	m_hq4x(160, 144),
	m_vram(new uint8_t[0x2000]),
	m_oam(new uint8_t[0x100]), // Should be 0x9F, but many access to 100
	m_reg(
		new GraphicsRegisters(
			core,
			m_oam,
			std::bind(&GameboyGraphics::backgroundChanged, this)
			)
		),
	m_backgroundRedraw(true),
	m_surface(surface),
	m_frameSkip(false),
	m_lastSkipped(false),
	m_stat_frames(0),
	m_stat_skipped(0),
	m_stat_redraw(0)
{
	MemoryMap& mapping(core.getMemoryMap());
	mapping.map(
		0x8000,
		0x2000,
		std::shared_ptr<VRAM>(
			new VRAM(
				m_vram,
				std::bind(&GameboyGraphics::backgroundChanged, this)
				)
			)
		);
	mapping.map(
		0xFE00,
		0x100,
		std::shared_ptr<VRAM>(
			new VRAM(
				m_oam,
				std::bind(&GameboyGraphics::oamChanged, this)
				)
			)
		);
	mapping.map(0xFF40, 12, m_reg);

	using namespace std::placeholders;
	m_mode0 = core.registerPeriodic(
		456, 252, std::bind(&glBoy::GameboyGraphics::periodicCallback, this, _1)
		);
	m_mode2 = core.registerPeriodic(
		456, 0, std::bind(&glBoy::GameboyGraphics::periodicCallback, this, _1)
		);
	m_mode3 = core.registerPeriodic(
		456, 80, std::bind(&glBoy::GameboyGraphics::periodicCallback, this, _1)
		);

	gettimeofday(&m_stat_fpsTimer, NULL);
}

GameboyGraphics::~GameboyGraphics()
{
	delete[] m_vram;
	delete[] m_oam;
}

void
GameboyGraphics::periodicCallback(Core::CallbackId id)
{
	// Mode 0: H-Blank
	// Mode 1: V-Blank
	// Mode 2: Reading from OAM
	// Mode 3: Reading from OAM and VRAM
	uint8_t mode;

	if (id == m_mode0)
	{
		renderScanLine();

		mode = (m_reg->LY > 143) ? 1 : 0;
	}
	else if (id == m_mode2)
	{
		startScanLine();
		mode = (m_reg->LY > 143) ? 1 : 2;
	}
	else // mode3
	{
		mode = (m_reg->LY > 143) ? 1 : 3;
	}

	m_reg->STAT = (m_reg->STAT & 0xF8) | mode;
}

void
GameboyGraphics::startScanLine()
{
	++m_reg->LY;
	if (m_reg->LY > 153) m_reg->LY = 0;

	m_lineScrollX[m_reg->LY] = m_reg->SCX;
	m_lineScrollY[m_reg->LY] = m_reg->SCY;

	bool equal(m_reg->LY == m_reg->LYC);
	if (equal)
	{
		m_reg->STAT |= 0x4;
	}
	else
	{
		m_reg->STAT &= 0xFB;
	}

	if ((m_reg->LY == 144))
	{
		// Draw the entire screen when entering vblank. This will be done just
		// under 60 times per second.
		if (!m_frameSkip || m_lastSkipped)
		{
			renderScreen();
			m_lastSkipped = false;
		}
		else
		{
			++m_stat_skipped;
			m_lastSkipped = true;
		}
		m_frameSkip = false;
		displayStats();

		//if (m_reg->STAT & 0x10) Seems to need to trigger regardless ?
		// Maybe we need to trigger 0x40 as well, depending on STAT ?
		{
			m_core.raiseInterrupt(0x40); // Start of vblank
		}
	}
	else if (
		(equal && (m_reg->STAT & 0x40)) || // Coincidence Interrupt
		(m_reg->STAT & 0x20) // OAM interrupt
		)
	{
		m_core.raiseInterrupt(0x48);
	}
}

void
GameboyGraphics::renderScanLine()
{
	// Entering H-Blank
	if (m_reg->STAT & 0x08 && (m_reg->LY <= 143))
	{
		m_core.raiseInterrupt(0x48);
	}

	// Don't bother drawing individual scanlines - just draw the entire screen
	// when entering VBLANK (which is actually done in the startScanLine method)
}

// Render the given tile to the screen at the given coordinates.
// The x and y may be outside the bounds of the screen, and may be negative.
// Only the visible part of the tile will be rendered.
void GameboyGraphics::renderTile(
	uint8_t tile,
	int x,
	int y,
	const FrameBuffer& frameBuffer
	)
{
	uint16_t tileAddr;
	if (m_reg->LCDC & 0x10)
	{
		tileAddr = 0x8000 + static_cast<uint16_t>(tile) * 16;
	}
	else
	{
		tileAddr = 0x9000 +
			(static_cast<int8_t>(tile) * static_cast<uint16_t>(16));
	}


	renderTile(tileAddr, x, y, frameBuffer, RENDER_BLIT, m_reg->BGP);
}

void GameboyGraphics::renderTile(
	uint16_t tileAddr,
	int x,
	int y,
	const FrameBuffer& frameBuffer,
	RenderMode mode,
	uint8_t palette,
	uint8_t sprite, // 0x80 -> 0xA8, 0 for background/window
	bool horizFlip,
	bool vertFlip,
	int tileHeight // 8 or 16
	)
{
	// Render an 8x8 or 8x16 tile.

	tileAddr -= 0x8000; // Use vram directly.

	if (vertFlip) tileAddr += (tileHeight - 1) * 2;
	int yStep = vertFlip ? -2 : 2; // Each row takes 2 bytes.
	if (y < 0) tileAddr += yStep * (-y);

	for (int posY = (y < 0 ? -y : 0);
		(posY < tileHeight) && (y + posY < frameBuffer.height);
		++posY, tileAddr += yStep
		)
	{
		uint8_t rowLSB(m_vram[tileAddr]);
		uint8_t rowMSB(m_vram[tileAddr + 1]);

		size_t pixelRowAddr((y + posY) * frameBuffer.width + x);

		int xStep = horizFlip ? 1 : -1;
		int xBit = horizFlip ? 0 : 7;
		if (x < 0) xBit += xStep * (-x);

		for (int posX = (x < 0 ? -x : 0);
			(posX < 8) && (x + posX < frameBuffer.width);
			++posX, xBit += xStep
			)
		{
			uint8_t col =
				(rowMSB >> xBit) & 0x01;
			col <<= 1;
			col |=
				(rowLSB >> xBit) & 0x01;

			Pixel& pixel = frameBuffer.frame[pixelRowAddr + posX];

			bool display;
			if (!sprite)
			{
				// Always display background and window data.
				// window data overwrites background data
				display = true;
			}
			else if (pixel.tileX != std::numeric_limits<int16_t>::max())
			{
				if (col)
				{
					if (x < pixel.tileX)
					{
						display = true;
					}
					else
					{
						// Existing pixel has priority. Note that transparent
						// areas of sprites are never output (ie. it remains
						// background), so there is no need to check of the
						// oldSprite was actually transparent.
						// Also note that we draw sprites in order, so this
						// sprite number can never be less than the already-
						// drawn sprite (and thus will not have priority)
						display = false;
					}
				}
				else
				{
					// Ignore transparent areas of the sprite.
					display = false;
				}
			}
			else
			{
				// Should this sprite pixel overwrite the background ?
				// Sprite colour 0 is transparent (shows background)
				// Background colour 0 is always "behind" the sprite.
				display =
					col &&
					((mode == RENDER_BLIT) || (pixel.colourIndex == 0));
			}

			if (display)
			{
				pixel.tileX = sprite ? x : std::numeric_limits<int16_t>::max();
				pixel.palette = palette;
				pixel.colourIndex = col;
			}
		}
	}
}

void
GameboyGraphics::renderScreen()
{
	FrameBuffer screenBuffer = { 160, 144, &m_screen[0][0]};

	if (m_reg->LCDC & 0x1)
	{
		// Draw background. Background tile map is a 256x256 virtual screen,
		// that is scrolled via SCX/SCY.

		if (m_backgroundRedraw)
		{
			m_backgroundRedraw = false;
			++m_stat_redraw;

			// Select one of two maps at 0x9800/0x9C00
			uint16_t tileMapAddr(m_reg->LCDC & 0x08 ? 0x1C00 : 0x1800);

			// Render entire tile map, then copy to the screen.
			memset(m_background, 0, sizeof(m_background));
			FrameBuffer buffer = { 256, 256, &m_background[0][0] };

			for (int y = 0; y < 32; ++y)
			{
				uint16_t yAddr(tileMapAddr + (y * 32));
				for (int x = 0; x < 32; ++x)
				{
					renderTile(
						m_vram[yAddr + x],
						x * 8,
						y * 8,
						buffer
						);
				}
			}
		}

		for (int y = 0; y < 144; ++y)
		{
			int scrollY = y + m_lineScrollY[y];
			if (scrollY > 255) scrollY -= 256;

			for (int x = 0; x < 160; ++x)
			{
				int scrollX = x + m_lineScrollX[y];
				if (scrollX > 255) scrollX -= 256;

				m_screen[y][x] = m_background[scrollY][scrollX];
			}
		}
	}
	else
	{
		// clear screen.
		memset(m_screen, 0, sizeof(Pixel) * 144 * 160);
	}


	if (m_reg->LCDC & 0x20)
	{
		// Draw window over top of existing background.

		// Draw background. Window tile map is a screen,
		// that is positioned via WX/WY.

		// Select one of two maps at 0x9800/0x9C00
		uint16_t tileMapAddr(m_reg->LCDC & 0x40 ? 0x1C00 : 0x1800);

		for (int y = 0; y < 18; ++y)
		{
			uint16_t yAddr(tileMapAddr + (y * 32));

			for (int x = 0; x < 20; ++x)
			{
				renderTile(
					m_vram[yAddr + x],
					x * 8 + m_reg->WX - 7,
					y * 8 + m_reg->WY,
					screenBuffer // Render directly to the screen.
					);
			}
		}
	}

	std::vector<uint8_t> sprite;
	sprite.resize(160);
	if (m_reg->LCDC & 0x2)
	{
		// Draw sprites

		for (int spriteNum = 0; spriteNum < 40; ++spriteNum)
		{
			Sprite sprite(getSprite(spriteNum));

			renderTile(
				0x8000 + static_cast<uint16_t>(sprite.tile) * 16,
				sprite.x,
				sprite.y,
				screenBuffer,
				(sprite.flags & 0x80) ? RENDER_SPRITE : RENDER_BLIT,
				(sprite.flags & 0x10) ? m_reg->OBP1 : m_reg->OBP0,
				static_cast<uint8_t>(spriteNum) | 0x80,
				(sprite.flags & 0x20) ? true : false, // horizontal flip
				(sprite.flags & 0x40) ? true : false, // vertical flip
				(m_reg->LCDC & 0x04) ? 16 : 8 // 8x8 or 8x16
				);
		}
	}
	updateOutputDevice();


}

GameboyGraphics::Sprite
GameboyGraphics::getSprite(int spriteNum) const
{
	Sprite result;

	result.y = static_cast<int>(m_oam[spriteNum * 4]) - 16;
	result.x = static_cast<int>(m_oam[spriteNum * 4 + 1]) - 8;
	result.tile = m_oam[spriteNum * 4 + 2];
	result.flags = m_oam[spriteNum * 4 + 3];

	return result;
}

void
GameboyGraphics::displayStats()
{
	m_stat_frames++;
	if (m_stat_frames == 300)
	{
		timeval startTime(m_stat_fpsTimer);
		gettimeofday(&m_stat_fpsTimer, 0);

		const double usec(1000000.0);
		double fps = static_cast<double>(m_stat_frames - m_stat_skipped) /
			(
				(m_stat_fpsTimer.tv_sec + m_stat_fpsTimer.tv_usec / usec) -
				(startTime.tv_sec + startTime.tv_usec / usec)
			);

		Log::Instance()(Log::DBG_STAT) <<
			"FPS = " << fps << " (error=" << (100 - (fps * 100 / 59.73)) << "%) "
			"backround redraw=" << m_stat_redraw << std::endl;
		m_stat_frames = 0;
		m_stat_skipped = 0;
		m_stat_redraw = 0;
	}
}

void
GameboyGraphics::blitGreyscale(uint8_t* dest)
{
	for (int y = 0; y < 144; ++y)
	{
		uint8_t* row(dest + y*160);
		for (int x = 0; x < 160; ++x)
		{
			int index(m_screen[y][x].colourIndex);
			int colour((m_screen[y][x].palette >> (index * 2)) & 0x3);
			row[x] = 255 - ( colour * 85);
		}
	}
}

void
GameboyGraphics:: updateOutputDevice()
{
	if (SDL_MUSTLOCK(m_surface))
	{
		if (SDL_LockSurface(m_surface) < 0)
		{
			return;
		}
	}

	// Use static buffers to reduce stack usage
	uint8_t textureData[144*160];
	blitGreyscale(&textureData[0]);

	// Allocate texture
	GLuint texture;
	glGenTextures(1, &texture);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture);

	// Set the texture's stretching properties
	// When using opengl shader scalers, we need to use nearest.
	uint32_t resample(
		(m_scaler == Scaler_None && m_resampler == Resample_Bilinear) ?
			GL_LINEAR : GL_NEAREST
		);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, resample);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, resample);

	// Create the texture image
	glTexImage2D(
		GL_TEXTURE_2D,
		0,
		GL_RGBA,
		160,
		144,
		0,
		GL_LUMINANCE,
		GL_UNSIGNED_BYTE,
		&textureData
		);

	m_hq4x.prepareShader(texture);

	// Output texture.

	glBegin(GL_QUADS);
		glTexCoord2i( 0, 0 ); glVertex2i( 0, 0 );
		glTexCoord2i( 1, 0 ); glVertex2i( 160, 0 );
		glTexCoord2i( 1, 1 ); glVertex2i( 160, 144 );
		glTexCoord2i( 0, 1 ); glVertex2i( 0, 144);
	glEnd();

	glLoadIdentity();
	SDL_GL_SwapBuffers();
	glDeleteTextures( 1, &texture );

	if (SDL_MUSTLOCK(m_surface))
	{
		SDL_UnlockSurface(m_surface);
	}
}

