//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.


#include "glBoy.hh"
#include "Core.hh"
#include "GameboyCart.hh"
#include "GameboyGraphics.hh"
#include "GameboyJoypad.hh"
#include "GameboySound.hh"
#include "GameboyTimer.hh"
#include "RAM.hh"
#include "ROM.hh"
#include "Log.hh"

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include <SDL/SDL_video.h>

#include <cassert>
#include <set>
#include <sstream>

#include <getopt.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

namespace
{
	SDL_Surface*
	initOpenGL(double xScale, double yScale, bool fullscreen)
	{
		glBoy::Log::Instance()(glBoy::Log::DBG_STAT) << "Enabling OpenGL" << std::endl;

		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

		unsigned int options(SDL_OPENGL);
		if (fullscreen)
		{
			options |= SDL_FULLSCREEN;
		}

		SDL_Surface* surface(
			SDL_SetVideoMode(160*xScale, 144*yScale, 32, options)
			);

		glEnable(GL_TEXTURE_2D);
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT);
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0f, 160, 144, 0.0f, -1.0f, 1.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		return surface;
	}

	void usage(const std::string& exe)
	{
		std::cout <<
			"Usage: " << exe << " "
				"[--scale ratio] [--width w] [--height h] \n"
				//"\t\t    [--resample-method={bilinear|nearest} \n"
				//"\t\t[--scaling-method={hq4x|scale2x|none}] "
				"rom.gb" << std::endl <<
			std::endl <<
				"\t--scale\t(Default=4.0)" << std::endl <<
					"\t\tMultiplier applied to the display output dimensions." << std::endl <<
				"\t--width" << std::endl <<
					"\t\tSet the display output dimensions directly" << std::endl <<
				"\t--height" << std::endl <<
					"\t\tSet the display output dimensions directly" << std::endl;
				/*"\t--scaling-method\n\t\t\t(Default=\"hq4x\")" << std::endl <<
					"\t\tSets the upscaling interpolation method." << std::endl <<
				"\t--resample-method\n\t\t\t(Default=\"bilinear\")" << std::endl <<
					"\t\tSets the resampling method." << std::endl;*/
	}
}

int main(int argc, char** argv)
{
	static struct option longOptions[] = 
	{
		{"video", 1, NULL, 'v'},
		{"scale", 1, NULL, 's'},
		{"width", 1, NULL, 'w'},
		{"height", 1, NULL, 'h'},
		//{"resample-method", 1, NULL, 'r'},
		//{"scaling-method", 1, NULL, 'm'},
		{NULL, 0, NULL, 0}
	};

	std::string filename("tetris.gb");
	glBoy::GameboyGraphics::PixelScaler scaler(glBoy::GameboyGraphics::Scaler_hq4x);
	glBoy::GameboyGraphics::ResampleMethod resampler(glBoy::GameboyGraphics::Resample_Bilinear);
	bool fullscreen(false);
	double scaleX(4.0);
	double scaleY(4.0);

	int optionIndex(0);
	int c;
	while (
		(c = getopt_long(argc, argv, "s:w:h:", &longOptions[0], &optionIndex))
			!= -1)
	{
		switch (c)
		{
			case 'h':
			{
				int height(160);
				std::stringstream convert(optarg);
				convert >> height;
				if (convert && height > 14 && height < 1440)
				{
					scaleY = height / 144.0;
				}
			}; break;
			case 'r':
			{
				if (std::string(optarg) == "bilinear")
				{
					resampler = glBoy::GameboyGraphics::Resample_Bilinear;
				}
				else
				{
					resampler = glBoy::GameboyGraphics::Resample_Nearest;
				}
			}; break;
			case 's':
			{
				double scale(1);
				std::stringstream convert(optarg);
				convert >> scale;
				if (convert && scale > 0.1 && scale < 10)
				{
					scaleX = scale;
					scaleY = scale;
				}
			}; break;
			case 'm':
			{
				if (std::string(optarg) == "hq4x")
				{
					scaler = glBoy::GameboyGraphics::Scaler_hq4x;
				}
				else if (std::string(optarg) == "scale2x")
				{
					scaler = glBoy::GameboyGraphics::Scaler_Scale2x;
				}
				else
				{
					scaler = glBoy::GameboyGraphics::Scaler_None;
				}
			}; break;
			case 'w':
			{
				int width(160);
				std::stringstream convert(optarg);
				convert >> width;
				if (convert && width > 16 && width < 1600)
				{
					scaleX = width / 160.0;
				}
			}; break;

			case '?':
				usage(argv[0]);
				exit(1);
				break;

			default:
				abort();
		};
	}

	if (optind < argc)
	{
		filename = argv[optind++];
	}

	glBoy::Core core;
	glBoy::MemoryMap& map(core.getMemoryMap());

	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_AUDIO);

	SDL_Surface* surface(initOpenGL(scaleX, scaleY, fullscreen));

	glBoy::GameboyGraphics graphics(core, surface, scaler, resampler);

	// Map the cart memory.
	int fd(open(filename.c_str(), O_RDONLY));
	assert(fd >= 0);

	struct stat buf;
	if (fstat(fd, &buf))
	{
		assert(!"Failed fstat");
	}
	uint8_t* cart = new uint8_t[buf.st_size];
	read(fd, cart, buf.st_size);
	close(fd);

	glBoy::GameboyCart tmp(core, cart, buf.st_size);


/*
	int biosFd(open("DMG_ROM.bin", O_RDONLY));
	uint8_t bios[256];
	read(biosFd, bios, 256);
	std::shared_ptr<glBoy::MemoryMap::Memory> biosMem(new glBoy::ROM(bios, 256));
	map.map(0, 256, biosMem);
*/

	// Map the cart mem.
	// Map the standard internal RAM, plus shadow.
	{
		std::shared_ptr<glBoy::MemoryMap::Memory> mem(new glBoy::RAM(8192));
		map.map(0xC000, 8192, mem);
		map.map(0xE000, 7680, mem);
	}
	// Map the "zero-page" high-speed internal ram.
	// Most interaction between GB hardware and the program occurs here.
	{
		std::shared_ptr<glBoy::MemoryMap::Memory> mem(new glBoy::RAM(128));
		map.map(0xFF80, 128, mem);
	}

	// Map the Interrupt flags. Currently handled in Core.cc
	{
		std::shared_ptr<glBoy::MemoryMap::Memory> mem(new glBoy::RAM(1));
		mem->write8(0, 0);
		map.map(0xFF0F, 1, mem);
	}

	// Map other devices
	std::shared_ptr<glBoy::GameboyJoypad> joypad(new glBoy::GameboyJoypad(core));
	map.map(0xFF00, 1, joypad);
	std::shared_ptr<glBoy::GameboyTimer> timer(
		new glBoy::GameboyTimer(
			core,
			std::bind(&glBoy::GameboyGraphics::setFrameSkip, &graphics),
			std::bind(&glBoy::GameboyGraphics::clearFrameSkip, &graphics)
			)
		);
	map.map(0xFF04, 4, timer);
	std::shared_ptr<glBoy::GameboySound> sound(new glBoy::GameboySound(core));
	map.map(0xFF10, 48, sound);

	// Cart starts from offset 0x100 with a NOP; JP
	core.getRegisters().PC = glBoy::htoz(0x100);
	core.getRegisters().SP = glBoy::htoz(0xFFFE);
	core.getRegisters().AF = glBoy::htoz(0x01B0);
	core.getRegisters().BC = glBoy::htoz(0x0013);
	core.getRegisters().DE = glBoy::htoz(0x00D8);
	core.getRegisters().HL = glBoy::htoz(0x014D);

	core.run();

	std::cerr << "Done" << std::endl;

	SDL_FreeSurface(surface);
	SDL_Quit();
}

