//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#include "glBoy.hh"
#include "MemoryMap.hh"
#include "Log.hh"
#include "util.hh"

#include <algorithm>
#include <cassert>

using namespace glBoy;

namespace
{
	class NullMemory : public MemoryMap::Memory
	{
	public:
		virtual uint8_t read8(uint16_t) { return 0; }
		virtual void write8(uint16_t, uint8_t) { }
	};

	struct NullDeleter
	{
		void operator()(MemoryMap::Memory*) {}
	};
}

MemoryMap::MemoryMap()
{
}

void
MemoryMap::map(
	uint16_t base,
	uint16_t size,
	const std::shared_ptr<Memory>& mem
	)
{
	MappedArea mapping(base, size, mem);

	MapType::iterator it(
		std::lower_bound(m_map.begin(), m_map.end(), mapping)
		);
	m_map.insert(it, mapping);

	Log::Instance()(Log::DBG_MEMORY) <<
		"Registering " << glBoy::hex(base) <<
		" for " << glBoy::hex(size) << " bytes."<< std::endl;
}

void
MemoryMap::remove(
	uint16_t base
	)
{
	MapType::iterator it(m_map.begin());
	for (; it != m_map.end() && it->base != base ; ++it) ;

	if (it != m_map.end())
	{
		Log::Instance()(Log::DBG_MEMORY) <<
			"Removing " << glBoy::hex(it->base) <<
			" for " << glBoy::hex(it->size) << " bytes."<< std::endl;

		m_map.erase(it);
	}
}


const MemoryMap::MappedArea&
MemoryMap::get(uint16_t address) const
{

/*
	MappedArea test(address, 0, std::shared_ptr<Memory>());
	MapType::const_iterator it(
		std::lower_bound(m_map.begin(), m_map.end(), test)
		);


	if (it != m_map.end() && it->base == address)
	{
		return *it;
	}
	else if (it != m_map.begin())
	{
		--it;
		if (address < it->base + static_cast<uint32_t>(it->size))
		{
			return *it;
		}
	}
*/


	for (
		MapType::const_iterator it(m_map.begin());
		it != m_map.end();
		++it
		)
	{
		if ((address >= it->base) &&
			(address < it->base + static_cast<uint32_t>(it->size))
			)
		{
			return *it;
		}
	}

	Log::Instance()(Log::DBG_MEMORY) <<
		"Returning NULL memory for address " << glBoy::hex(address) << std::endl;

	static NullMemory s_null;
	static MappedArea s_nullArea(
		0, 0, std::shared_ptr<Memory>(&s_null, NullDeleter())
		);
	return s_nullArea;
}

uint8_t
MemoryMap::read8(DWORD address, int8_t offset)
{
	uint16_t offsetAddress(address.host() + offset);
	const MappedArea& mapping(get(offsetAddress));
	return mapping.memory->read8(offsetAddress - mapping.base);
}

DWORD
MemoryMap::read16(DWORD address, int8_t offset)
{
	uint16_t offsetAddress(address.host() + offset);

	const MappedArea& loMap(get(offsetAddress));
	uint8_t lo(loMap.memory->read8(offsetAddress - loMap.base));

	++offsetAddress;
	const MappedArea& hiMap(get(offsetAddress));
	uint16_t hi(hiMap.memory->read8(offsetAddress - hiMap.base));

	return htoz((hi << 8) | lo);
}

void
MemoryMap::write8(DWORD address, uint8_t value)
{
	write8(address, 0, value);
}

void
MemoryMap::write8(DWORD address, int8_t offset, uint8_t value)
{
	uint16_t offsetAddress(address.host() + offset);
	const MappedArea& mapping(get(offsetAddress));
	mapping.memory->write8(offsetAddress - mapping.base, value);
}

void
MemoryMap::write16(DWORD address, DWORD value)
{
	write16(address, 0, value);
}

void
MemoryMap::write16(DWORD address, int8_t offset, DWORD value)
{
	uint16_t offsetAddress(address.host() + offset);
	const MappedArea& loMap(get(offsetAddress));
	loMap.memory->write8(offsetAddress - loMap.base, value.l);

	++offsetAddress;
	const MappedArea& hiMap(get(offsetAddress));
	hiMap.memory->write8(offsetAddress - hiMap.base, value.h);
}

void
MemoryMap::copy(uint8_t* dest, uint16_t srcAddress, int bytes)
{
	const MappedArea& srcMap(get(srcAddress));

	assert((srcAddress + bytes) <= (srcMap.base + srcMap.size));

	srcMap.memory->copy(dest, srcAddress - srcMap.base, bytes);
}

void
MemoryMap::Memory::copy(uint8_t* dest, uint16_t address, int bytes)
{
	for (int i = 0; i < bytes; ++i)
	{
		dest[i] = read8(address + i);
	}
}

