//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#ifndef GLBOY_Registers_hh
#define GLBOY_Registers_hh

#include "glBoy.hh"
#include "DWORD.hh"

namespace glBoy
{

class Core;

/** 8-bit packed representation of the Z80 flags register.
 *
 *   C    Carry.  LSB of Flag register
 *   N    Add/Subtract
 *   P    (V) Parity/Overflow
 *   U3   3rd bit of last 8bit op that altered flags
 *   H    Half-Carry (BCD)
 *   U5   5th bit of last 8bit op that altered flags
 *   Z    Zero Flag
 *   S    Sign Flag. MSB of Flag register
 */
struct __attribute__ ((__packed__)) Flags
{
	union
	{
		struct __attribute__ ((__packed__))
		{
			unsigned C:1; /// Carry.  LSB of Flag register
			unsigned N:1; /// Add/Subtract
			unsigned P:1; /// (V) Parity/Overflow
			unsigned U3:1; /// 3rd bit of last 8bit op that altered flags
			unsigned H:1; /// Half-Carry (BCD)
			unsigned U5:1; /// 5th bit of last 8bit op that altered flags
			unsigned Z:1; /// Zero Flag
			unsigned S:1; /// Sign Flag. MSB of Flag register
		};
		reg8_t byte;
	};

	// No ctor allowed for POD.
	void reset() { byte = 0; }

	// Update the flags based on an operands value (eg. for simple assignment)
	void set(op8_t operand, const Core& context);

	// Update the flags when changing BC
	void setCounter(DWORD BC);

	// Set from A - B
	void setSub(op8_t A, op8_t B, DWORD BC);
};

/** Z80 register access.
 *
 * Registers may be accessed in either 8-bit (eg. A) 
 * or by their 16-bit pair (eg. AF)
 *
 */
struct Registers
{
	// No ctor allowed for POD.
	void reset();

#define Z80_REG_STRUCT(h,l) \
	struct \
	{ \
		l; \
		h; \
	};

	// Note: The order of these unions/structs is important, as
	// offsets are taken into the Register struct
	reg8_t begin8[0];
	DWORD begin16[0];

	union
	{
		Z80_REG_STRUCT(reg8_t B, reg8_t C);
		DWORD BC;
	};

	union
	{
		Z80_REG_STRUCT(reg8_t D, reg8_t E);
		DWORD DE;
	};

	union
	{
		Z80_REG_STRUCT(reg8_t H, reg8_t L);
		DWORD HL;
	};

	union
	{
		Z80_REG_STRUCT(reg8_t A, Flags F);
		DWORD AF;
	};


#undef Z80_REG_STRUCT

	// Special Purpose
	reg8_t I;      // Interrupt Vector
	reg8_t R;      // Memory Refresh
	DWORD IX;     // Index Register
	DWORD IY;     // Index Register
	DWORD SP;     // Stack Pointer
	DWORD PC;     // Program Counter
};

} // namespace glBoy

#endif

