//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#include "ALU.hh"
#include "DAA.hh"

#include "stddef.h"

#include <cassert>
#include <iostream>
#include <ios>

using namespace glBoy;

void add(uint8_t a, uint8_t b)
{
	Flags f;
	f.reset();
	uint8_t result(ALU::add(a, b, f));

	uint16_t testFlags;
	uint8_t testResult;
	__asm__ (
		"movb %2, %%al\n\t"
		"addb %3, %%al\n\t"
		"pushf\n\t"
		"mov %%al, %1\n\t"
		"pop %%ax\n\t"
		"movw %%ax, %0\n\t"
		: "=g" (testFlags), "=g" (testResult)
		: "g" (a), "g" (b)
		: "eax", "cc"
		);

	assert(result == testResult);
	assert(f.S == (testFlags & 0x80) >> 7);
	assert(f.Z == (testFlags & 0x40) >> 6);
	assert(f.H == (testFlags & 0x10) >> 4);
	assert(f.P == (testFlags & 0x800) >> 11); // 0x4 for parity
	assert(f.C == (testFlags & 0x1));
}

void sub(uint8_t a, uint8_t b)
{
	Flags f;
	f.reset();
	uint8_t result(ALU::sub(a, b, f));

	uint16_t testFlags;
	uint8_t testResult;
	__asm__ (
		"movb %2, %%al\n\t"
		"subb %3, %%al\n\t"
		"pushf\n\t"
		"mov %%al, %1\n\t"
		"pop %%ax\n\t"
		"movw %%ax, %0\n\t"
		: "=g" (testFlags), "=g" (testResult)
		: "g" (a), "g" (b)
		: "eax", "cc"
		);

	assert(result == testResult);
	assert(f.S == (testFlags & 0x80) >> 7);
	assert(f.Z == (testFlags & 0x40) >> 6);
	assert(f.H == (testFlags & 0x10) >> 4);
	assert(f.P == (testFlags & 0x800) >> 11); // 0x4 for parity
	assert(f.C == (testFlags & 0x1));
}

void bitwiseAnd(uint8_t a, uint8_t b)
{
	Flags f;
	f.reset();
	uint8_t result(ALU::bitwiseAnd(a, b, f));

	uint16_t testFlags;
	uint8_t testResult;
	__asm__ (
		"movb %2, %%al\n\t"
		"andb %3, %%al\n\t"
		"pushf\n\t"
		"mov %%al, %1\n\t"
		"pop %%ax\n\t"
		"movw %%ax, %0\n\t"
		: "=g" (testFlags), "=g" (testResult)
		: "g" (a), "g" (b)
		: "eax", "cc"
		);

	assert(result == testResult);
	assert(f.S == (testFlags & 0x80) >> 7);
	assert(f.Z == (testFlags & 0x40) >> 6);
	//assert(f.H == (testFlags & 0x10) >> 4);
	// H is always set for AND for Z80, but not X86
	assert(f.H);
	assert(f.P == (testFlags & 0x04) >> 2); // 0x800 for overflow
	assert(f.C == (testFlags & 0x1));
}

void bitwiseOr(uint8_t a, uint8_t b)
{
	Flags f;
	f.reset();
	uint8_t result(ALU::bitwiseOr(a, b, f));

	uint16_t testFlags;
	uint8_t testResult;
	__asm__ __volatile__ (
		"movb %2, %%al\n\t"
		"orb %3, %%al\n\t"
		"pushf\n\t"
		"mov %%al, %1\n\t"
		"pop %%ax\n\t"
		"movw %%ax, %0\n\t"
		: "=g" (testFlags), "=g" (testResult)
		: "g" (a), "g" (b)
		: "eax", "cc"
		);

	assert(result == testResult);
	assert(f.S == (testFlags & 0x80) >> 7);
	assert(f.Z == (testFlags & 0x40) >> 6);
	assert(f.H == (testFlags & 0x10) >> 4);
	assert(f.P == (testFlags & 0x04) >> 2); // 0x800 for overflow
	assert(f.C == (testFlags & 0x1));
}

void bitwiseXor(uint8_t a, uint8_t b)
{
	Flags f;
	f.reset();
	uint8_t result(ALU::bitwiseXor(a, b, f));

	uint16_t testFlags;
	uint8_t testResult;
	__asm__ __volatile__(
		"movb %2, %%al\n\t"
		"xorb %3, %%al\n\t"
		"pushf\n\t"
		"mov %%al, %1\n\t"
		"pop %%ax\n\t"
		"movw %%ax, %0\n\t"
		: "=g" (testFlags), "=g" (testResult)
		: "g" (a), "g" (b)
		: "eax", "cc"
		);

	assert(result == testResult);
	assert(f.S == (testFlags & 0x80) >> 7);
	assert(f.Z == (testFlags & 0x40) >> 6);
	assert(f.H == (testFlags & 0x10) >> 4);
	assert(f.P == (testFlags & 0x04) >> 2); // 0x800 for overflow
	assert(f.C == (testFlags & 0x1));
}

uint8_t daa(uint8_t a, uint8_t b)
{
	Flags f;
	f.reset();
	uint8_t tmp(ALU::add(a, b, f));

	return DAA(tmp, f);

}

uint8_t daaSub(uint8_t a, uint8_t b)
{
	Flags f;
	f.reset();
	uint8_t tmp(ALU::sub(a, b, f));

	return DAA(tmp, f);

}

int main()
{
	// Ensure packed representations are the expected size,
	// and usable within union types.
	assert(sizeof(DWORD) == 2);
	assert(sizeof(Flags) == 1);

	// Ensure 16bit registers are aligned to 16-bit boundaries.
	assert(offsetof(Registers, BC) == 0);
	assert(offsetof(Registers, AF) == 6);

	add(0, 0);
	add(1, 1);
	add(-1, 1);
	add(-128, 127);
	add(127, 1);
	add(-128, -1);
	add(5, 5);
	add(15, 1);
	add(15, 128);

	Flags f;
	f.reset();
	assert(ALU::add(htoz(0x1000), htoz(0x5555), f).host() == 0x6555);
	assert(!f.C);
	assert(ALU::add(htoz(0x1000), htoz(0xF555), f).host() == 0x0555);
	assert(f.C);


	sub(0, 0);
	sub(1, 1);
	sub(-1, 1);
	sub(-128, 127);
	sub(127, 1);
	sub(-128, -1);
	sub(5, 5);
	sub(15, 1);
	sub(15, 128);
	sub(0, 0xb);

	bitwiseAnd(0, 0);
	bitwiseAnd(1, 1);
	bitwiseAnd(-1, 1);
	bitwiseAnd(-128, 127);
	bitwiseAnd(127, 1);
	bitwiseAnd(-128, -1);
	bitwiseAnd(5, 5);
	bitwiseAnd(15, 1);
	bitwiseAnd(15, 128);

	bitwiseOr(0, 0);
	bitwiseOr(1, 1);
	bitwiseOr(-1, 1);
	bitwiseOr(-128, 127);
	bitwiseOr(127, 1);
	bitwiseOr(-128, -1);
	bitwiseOr(5, 5);
	bitwiseOr(15, 1);
	bitwiseOr(15, 128);

	bitwiseXor(0, 0);
	bitwiseXor(1, 1);
	bitwiseXor(-1, 1);
	bitwiseXor(-128, 127);
	bitwiseXor(127, 1);
	bitwiseXor(-128, -1);
	bitwiseXor(5, 5);
	bitwiseXor(15, 1);
	bitwiseXor(15, 128);

	assert(daa(0x01, 0x01) == 0x02);
	assert(daa(0x01, 0x09) == 0x10);
	assert(daa(0x01, 0x10) == 0x11);
	assert(daa(0x09, 0x09) == 0x18);
	assert(daa(0x50, 0x49) == 0x99);
	assert(daa(0x99, 0x0) == 0x99);
	assert(daa(0x27, 0x27) == 0x54);
	assert(daa(0x90, 0x10) == 0x0); // Carry from 0x100
	assert(daa(0x99, 0x99) == 0x98); // Carry from 0x198

	assert(daaSub(0x10, 0x01) == 0x09);
	assert(daaSub(0x10, 0x20) == 0x90);

	std::cout << "Tests complete" << std::endl;
}
