#!/usr/bin/perl -w
#	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
#
#	This file is part of glBoy.
#
#	glBoy is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	glBoy is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

use warnings;
use strict;
use Carp;

use XML::LibXML::Reader;

# < Ugly Globals >
my %RegMasks = ();
my %BitMasks = ();
my %Instructions = ();

my $VARIANT = "gb80";

# </ Ugly Globals >

my $reader = new XML::LibXML::Reader(location => "InstructionSet.opcode")
	|| die "Cannot read opcode file\n";

open (TABLES, '>InstructionSet_Tables.cc') || die "Could not write to file\n";
open (CODE, '>InstructionSet_Opcodes.cc') || die "Could not write to file\n";

$reader->read;
$reader->name eq "opcodes" || die "Invalid xml\n";

my $node = $reader->read;
while ($node)
{
	if ($reader->nodeType == XML_READER_TYPE_ELEMENT)
	{
		# Convert subtree to a DOM object.
		my $subtree = $reader->copyCurrentNode(1);

		if ($reader->name eq "registerMask")
		{
			processRegisterMask($subtree);
		}
		elsif ($reader->name eq "bitMask")
		{
			processBitMask($subtree);
		}
		elsif ($reader->name eq "instruction")
		{
			processInstruction($subtree)
		}
		else
		{
			# Ignore unknown XML
		}
	}

	$node = $reader->next();
}

outputInstructionTable();

close TABLES;
close CODE;

sub outputInstructionTable
{
	foreach my $prefix (keys %Instructions)
	{
		my $outPrefix = "OpcodeTable$prefix";
		$outPrefix =~ s/0x//;
		$outPrefix =~ s/default//;
		print TABLES "static void* $outPrefix\[256\] =\n{\n";

		my %opcodes = %{$Instructions{$prefix}};

		my $count = 0;
		while ($count < 256)
		{
			if ($Instructions{$prefix}{$count})
			{
				print TABLES "&&$Instructions{$prefix}{$count}";
			}
			else
			{
				print TABLES "&&BAD_OPCODE";
			}

			$count = $count + 1;

			if ($count != 256)
			{
				print TABLES ",\t";
			}
			if ($count % 4 == 0)
			{
				print TABLES "\n";
			}
		}
		print TABLES "};\n\n";
	}
}


sub processRegisterMask
{
	my $node = shift;

	my $name = $node->findvalue('./@name') || die;
	my $maskBits = $node->findvalue('./@maskBits') || die;
	($maskBits > 0 && $maskBits < 8) || die;
	my $regBytes = $node->findvalue('./@regBytes') || die;
	($regBytes > 0 && $regBytes <= 2) || die;

	my $variant = $node->findvalue('./@variant');

	if ($variant && $variant ne $VARIANT)
	{
		return;
	}

	my $max = 1 << $maskBits;
	my $count = 0;

	print TABLES "static size_t regMask_$name\[$max\] =\n{\n";

	my @masks = ();

	while ($count < $max)
	{
		my $mask = sprintf("%0$maskBits"."b", $count);
		my $reg = $node->findvalue("./reg[\@mask='$mask']/\@name");

		if ($count != 0) { print TABLES ",\n"; }
		if ($reg && length $reg != 0)
		{
			print TABLES "\toffsetof(glBoy::Registers, $reg) / $regBytes";
			push @masks, $mask;
		}
		else
		{
			print TABLES "\tsize_t(-1)";
		}

		$count = $count + 1;
	}
	print TABLES "\n};\n\n";

	$RegMasks{$name}{'masks'} = \@masks;
	$RegMasks{$name}{'maskBits'} = $maskBits;
	$RegMasks{$name}{'regBytes'} = $regBytes;

}

sub processBitMask
{
	my $node = shift;

	my $name = $node->findvalue('./@name') || die;
	my $maskBits = $node->findvalue('./@maskBits') || die;
	($maskBits > 0 && $maskBits <= 8) || die;

	my $max = 1 << $maskBits;
	my $count = 0;

	my @masks = ();

	while ($count < $max)
	{
		my $mask = sprintf("%0$maskBits"."b", $count);
		push @masks, $mask;
		$count = $count + 1;
	}

	$BitMasks{$name}{'masks'} = \@masks;
	$BitMasks{$name}{'maskBits'} = $maskBits;
}

sub expandMask
{
	my $mask = shift;

	my @result = ();

	if ($mask =~ /([a-zA-Z])/)
	{
		my $match = $1;

		if ($BitMasks{$match})
		{
			foreach my $bits (@{$BitMasks{$match}{'masks'}})
			{
				my $nextMask = $mask;
				$nextMask =~ s/$match/$bits/;
				push @result, expandMask($nextMask);
			}
		}
		elsif ($RegMasks{$match})
		{
			foreach my $bits (@{$RegMasks{$match}{'masks'}})
			{
				my $nextMask = $mask;
				$nextMask =~ s/$match/$bits/;
				push @result, expandMask($nextMask);
			}

		}
		$mask =~ s/$match/_/;
	}
	else
	{
		my $pos = 0;
		my $value = 0;

		if (length $mask != 8) { croak "Bad Mask $mask\n"; }
		while ($pos < 8)
		{
			my $char = substr($mask, $pos, 1);
			if ($char == 1)
			{
				$value = $value + (1 << (7 - $pos));
			}
			$pos = $pos + 1;
		}
		push @result, $value;
	}
	return @result;
}

sub createRegReference
{
	my $mask = shift;

	my %seenChars = ();

	my @dbgResult = ();

	# This is probably not the fastest way, but who cares.
	my $pos = 0;
	my $bitPos = 0;
	while ($pos < length $mask)
	{
		my $char = substr($mask, $pos, 1);
		if ($char =~ /[a-zA-Z]/ && $RegMasks{$char})
		{
			my $outputName = $char;
			if ($seenChars{$outputName})
			{
				$seenChars{$outputName} = $seenChars{$outputName} + 1;
				$outputName .= $seenChars{$outputName};
			}
			else
			{
				$seenChars{$outputName} = 1;
			}

			push @dbgResult, $outputName;

			my $maskBits = $RegMasks{$char}{'maskBits'};
			my $shiftNum = 8 - ($bitPos + $RegMasks{$char}{'maskBits'});
			if ($RegMasks{$char}{'regBytes'} == 1)
			{
				print CODE "\tuint8_t& $outputName(m_reg.begin8[regMask_$char\[(opcode >> $shiftNum) & (uint8_t(-1) >> (8 - $maskBits)) \]\]);\n";
			}
			else
			{
				print CODE "\tDWORD& $outputName(m_reg.begin16[regMask_$char\[(opcode >> $shiftNum) & (uint8_t(-1) >> (8 - $maskBits)) \]\]);\n";
			}

			$bitPos = $bitPos + $maskBits;
		}
		elsif ($char =~ /[a-zA-Z]/ && $BitMasks{$char})
		{
			my $maskBits = $BitMasks{$char}{'maskBits'};
			my $shiftNum = 8 - ($bitPos + $BitMasks{$char}{'maskBits'});

			print CODE "uint8_t $char((opcode >> $shiftNum) & (uint8_t(-1) >> (8 - $maskBits)));\n";
			$bitPos = $bitPos + $maskBits;
		}
		else
		{
			$bitPos = $bitPos + 1;
		}

		$pos = $pos + 1;
	}
	return @dbgResult;
}

sub processInstruction
{
	my $node = shift;

	my $name = $node->findvalue('./@name') || die;
	my $clock = $node->findvalue('./@clock');
	my $mask = $node->findvalue('./opcode/@mask');
	my $prefix = $node->findvalue('./opcode/@prefix');
	my $operand = $node->findvalue('./opcode/@operand');
	my $signedOperand = $node->findvalue('./opcode/@signed_operand');
	my $variant = $node->findvalue('./@variant');

	if ($variant && $variant ne $VARIANT)
	{
		return;
	}
	if ($VARIANT eq "gb80" && $prefix && $prefix ne "0xCB")
	{
		return;
	}

	print CODE "// Instruction: $name\n";
	print CODE "// Clock: $clock\n";
	my $label = $name;
	$label =~ tr/ ,'+()/______/;
	print CODE "$label:\n";
	print CODE "{\n";


	# Note: May have both signed operand (index offset) and operand
	if ($signedOperand && length $signedOperand == 1)
	{
		print CODE "\tint8_t $signedOperand(op8());\n";
	}

	if ($operand && length $operand == 1)
	{
		if (length($prefix) > 4)
		{
			print CODE "\t uint8_t& $operand(doublePrefixOperand);\n";
		}
		elsif (uc($operand) eq $operand)
		{
			print CODE "\tDWORD $operand(op16());\n";
		}
		else
		{
			print CODE "\tuint8_t $operand(op8());\n";
		}
	}

	my @dbgRegs = createRegReference($mask);

	print CODE "#ifdef GLBOY_DEBUG\n";
	print CODE "std::ostream& trace(log(Log::DBG_TRACE));\n".
		"trace << ".
		"opcodePC << \": \";\n".
		"if (prefix) trace << hex(prefix) << \" \";\n".
		"trace << hex(opcode) << \" $name\" << ".
		($operand ? "\" ($operand=\" << $operand << \")\" << " : "") .
		($dbgRegs[0] ? "\" ($dbgRegs[0]=\" << $dbgRegs[0] << \")\" << " : "") .
		"std::endl;\n";
	print CODE "#endif\n";

	my $statements = $node->findvalue('./text()');
	chomp($statements);
	print CODE $statements."\n";

	print CODE "\tclock += $clock;\n";

	print "Processing $name\n";
	my @opcodes = expandMask($mask);

	if (!$prefix || length $prefix == 0)
	{
		$prefix = "default";
	}

	foreach my $opcode (@opcodes)
	{
		if ($Instructions{$prefix}{$opcode})
		{
			die "Duplicate opcode. Prefix = $prefix. Mask = $mask, $name.\n".
				"Conflicting label = $Instructions{$prefix}{$opcode}\n";
		}
		$Instructions{$prefix}{$opcode} = $label;
	}

	print CODE "\n}\n";
	print CODE "goto END_INSTRUCTIONS;\n";
	print CODE "\n\n";
}

